// -*-c++-*-

/*!
  \file kick_table.cpp
  \brief kick info holder Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <algorithm>
#include <functional>

#include "generic_utils/logger.h"
#include "soccer_param/server_param.h"
#include "soccer_param/player_type.h"

#include "world_model.h"
#include "kick_state.h"

namespace rcsc {

namespace {

const double POWER_THR_RATE = 0.9;

}

const int KickPoint::MAX_INTERFERE = 4;

/*-------------------------------------------------------------------*/

struct KickPointPtrCmp
    : public std::binary_function< const KickPoint*,
                                   const KickPoint*,
                                   bool > {
public:
    result_type operator()( first_argument_type lhs,
                            second_argument_type rhs ) const
      {
          return lhs->interfereCycle() > rhs->interfereCycle();
      }
};


/*-------------------------------------------------------------------*/
/*!

*/
KickTable::KickTable()
{
    for ( int i = 0; i < MAX_DEPTH; ++i )
    {
        M_kick_points[i].reserve( ANGLE_DIVS_HIGH * 2 );
        M_kick_points_ptr[i].reserve( ANGLE_DIVS_HIGH * 2 );
    }
}

/*-------------------------------------------------------------------*/
/*!
  \param self_pos self global coordinates as the base point
  \param self_body self body angle to calculate the kick rate
*/
void
KickTable::update( const WorldModel & world )
{
    if ( world.self().isKickable() )
    {
        for ( int i = 0; i < MAX_DEPTH; ++i )
        {
            M_kick_points[i].clear();
            M_kick_points_ptr[i].clear();
        }

        // set ball point and kick rate
        setPoints( world );
        //! sort kick points by interfere cycles
        sortPoints();
        dlog.addln( Logger::DEBUG_MULTI_KICK,
                    "   kick_points [0].size=%d [1].size=%d",
                    M_kick_points[0].size(), M_kick_points[1].size() );
    }
}

/*-------------------------------------------------------------------*/
/*!
  \brief create all possible subtargets.
  if opponents are not near, sparse subtargets.
  else, more angle division
*/
void
KickTable::setPoints( const WorldModel & world )
{
    const SelfObject & self_o = world.self();
    const PlayerType & t = self_o.playerType();

    // default_type: 0.3 + (0.7 + 0.085)*0.4 = 0.614: col_buf = 0.229
    // max_type:     0.3 + (0.9 + 0.085)*0.4 = 0.694: col_buf = 0.309
    const double dist1 = ( type.playerSize()
                           + ( type.kickableMargin()
                               + ServerParam::i().ballSize() ) * 0.4 );
    // default_type: 0.3 + (0.7 + 0.085)*0.7 = 0.8495: area_buf = 0.2355
    // max_type:     0.3 + (0.9 + 0.085)*0.7 = 0.9895: area_buf = 0.2955
    double tmpd = ( type.playerSize()
                    + ( type.kickableMargin()
                        + ServerParam::i().ballSize() ) * 0.7 );
    const double dist2 = std::floor( tmpd * 10.0 ) * 0.1;


    ////////////////////////////////////////////////////////////
    // decide subtarget frequency
    bool exist_near_opp = false;
    if ( ! world.opponentsNearSelf().empty()
         && world.opponentsNearSelf().front()->distFromSelf() <
         ( ServerParam::i().defaultKickableArea() + 0.2 ) * 2.0 )
    {
        exist_near_opp = true;
    }

    const int ANGLE_DIVS = ( exist_near_opp
                             ? ANGLE_DIVS_HIGH
                             : ANGLE_DIVS_LOW );
    const double dir_inc = 360.0 / ANGLE_DIVS;

    ////////////////////////////////////////////////////////////
    int id = 0;
    AngleDeg ball_angle = -180.0;
    for ( int i = 0; i < ANGLE_DIVS; ++i )
    {
        const double dir_diff = ( self_o.body() - ball_angle ).abs();

        Vector2D brpos1( 1, dist1, ball_angle );
        Vector2D brpos2 = brpos1;
        brpos2 /= dist1;
        brpos2 *= dist2;

        double krate1 = type.kickRate( dist1, dir_diff );
        double krate2 = type.kickRate( dist2, dir_diff );

        Vector2D spos = self_o.pos();
        Vector2D svel = self_o.vel();
        for ( int depth = 0; depth < MAX_DEPTH; ++depth )
        {
            spos += svel;
            svel *= type.playerDecay();
            // register dist1 point
            Vector2D tpos = spos + brpos1;
            int icyc = calcInterfereCycle( world, tpos, depth );
            if ( icyc > 0 )
            {
#ifdef DEBUG
                dlog.addln( Logger::DEBUG_MULTI_KICK,
                            "   kick_state[%d:%d](1) (%f %f) id=%d angle=%f kick_rate=%f, icyc=%d",
                            i, depth,
                            tpos.x, tpos.y, id, ball_angle.degree(), krate1, icyc );
#endif
                icyc = std::min( KickPoint::MAX_INTERFERE, icyc );
                M_kick_points[depth].push_back( KickPoint( id++, tpos, krate1,
                                                           icyc ) );
            }
            // register dist2 point
            tpos = spos + brpos2;
            icyc = calcInterfereCycle( world, tpos, depth );
            if ( icyc > 0 )
            {
#ifdef DEBUG
                dlog.addln( Logger::DEBUG_MULTI_KICK,
                            "   kick_state[%d:%d](2) (%f %f) id=%d angle=%f kick_rate=%f, icyc=%d",
                            i, depth,
                            tpos.x, tpos.y, id, ball_angle.degree(), krate2, icyc );
#endif
                icyc = std::min( KickPoint::MAX_INTERFERE, icyc );
                M_kick_points[depth].push_back( KickPoint( id++, tpos, krate2,
                                                           icyc ) );
            }
        }

        ball_angle += dir_inc;
    }

    // copy to the const_pointer container
    for ( int depth = 0; depth < MAX_DEPTH; ++depth )
    {
        const KickPointCont::iterator it_end = M_kick_points[depth].end();
        for ( KickPointCont::iterator it = M_kick_points[depth].begin();
              it != it_end;
              ++it )
        {
            M_kick_points_ptr[depth].push_back( &( *it ) );
        }
    }
}

/*-------------------------------------------------------------------*/
/*!
  update each KickPoint's interfere_cycle
*/
int
KickTable::calcInterfereCycle( const WorldModel& world,
                               const Vector2D& point,
                               const int depth )
{
    // the absolute coordinates are over pitch area.
    if ( point.absX() > ServerParam::i().pitchHalfLength() - 0.3
         || point.absY() > ServerParam::i().pitchHalfWidth() - 0.3 )
    {
        return 0;
    }

    int cycle = 100;
    PlayerPtrCont::const_iterator it_end = world.getOpponentsFromBall().end();
    for ( PlayerPtrCont::const_iterator it = world.getOpponentsFromBall().begin();
          it != it_end;
          ++it )
    {
        if ( (*it)->posCount() > 5 ) // Magic Number
        {
            continue;
        }
        if ( (*it)->distFromBall() > 6.0 ) // Magic Number
        {
            break;
        }

        // maximal accuracy count is "2"
        // if posCount() is over 2, count buffer is normalized under 2.
        int tmp_cyc = (*it)->predictReachCycle( point, depth + 1, 2 );

        if ( tmp_cyc <= 0 )
        {
            return 0;
        }

        if ( tmp_cyc < cycle )
        {
            cycle = tmp_cyc;
        }
    }

    return cycle;
}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickTable::predictReachCycle( const PlayerObject & player,
                              const Vector2D & point,
                              const int wait_step,
                              const int max_step )
{



}

/*-------------------------------------------------------------------*/
/*!

*/
void
KickTable::sortPoints()
{
    for ( int i = 0; i < MAX_DEPTH; ++i )
    {
        std::sort( M_kick_points_ptr[i].begin(),
                   M_kick_points_ptr[i].end(),
                   KickPointPtrCmp() );
    }
}

}
