// -*-c++-*-

/*!
  \file bhv_mark_player.cpp
  \brief do 1-on-1 defense
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_mark_player.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/abstract_player_object.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>
#if 0
#include <rcsc/action/body_intercept.h>
#else
#include "body_intercept_devel.h"
#endif
#include "body_go_to_point_devel.h"
#include <rcsc/action/body_turn_to_angle.h>
#include <rcsc/action/neck_turn_to_player_or_scan.h>

static
rcsc::Vector2D
s_get_position( rcsc::PlayerAgent * agent,
                const rcsc::Vector2D & ball_pos,
                double base_dist,
                const rcsc::Vector2D & self_pos,
                const rcsc::AbstractPlayerObject * target );

/*-------------------------------------------------------------------*/
/*!

*/
Bhv_MarkPlayer::Bhv_MarkPlayer( const rcsc::AbstractPlayerObject * target )
    : M_target( target )
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Bhv_MarkPlayer::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": mark player, target unum = %d",
                        M_target->unum() );
    agent->debugClient().addMessage( "Mark" );

    const rcsc::WorldModel & wm = agent->world();

    const rcsc::Vector2D ball_pos = wm.ball().inertiaPoint( 1 );
    const rcsc::Vector2D target_player_pos = M_target->inertiaPoint( 1 );

    double base_dist = std::min( 18.0,
                                 (target_player_pos
                                  - rcsc::ServerParam::i().ourTeamGoalPos()).r() );

    rcsc::Vector2D pos = s_get_position( agent,
                                         target_player_pos,
                                         base_dist,
                                         wm.self().pos(),
                                         M_target );
    agent->debugClient().addLine( wm.self().pos(), target_player_pos );
    agent->debugClient().addLine( wm.self().pos(), pos );

    agent->setNeckAction( new rcsc::Neck_TurnToPlayerOrScan( M_target ) );

    if ( Body_GoToPoint_Devel
           ( pos,
             2.0,
             rcsc::ServerParam::i().maxPower() ).execute( agent )
        || Body_GoToPoint_Devel
           ( pos,
             1.0,
             rcsc::ServerParam::i().maxPower() ).execute( agent )
        || Body_GoToPoint_Devel
           ( pos,
             0.5,
             rcsc::ServerParam::i().maxPower() ).execute( agent )
        || Body_GoToPoint_Devel
           ( pos,
             0.1,
             rcsc::ServerParam::i().maxPower() ).execute( agent ) )
    {
        return true;
    }


    const rcsc::Vector2D final_body_pos = wm.self().inertiaFinalPoint();

    rcsc::AngleDeg diff_body_angle = 0.0;

    if ( (ball_pos - final_body_pos).r() <= 10.0
         && (wm.existKickableOpponent() || wm.existKickableTeammate()) )
    {
        if ( final_body_pos.y > 0.0 )
        {
            diff_body_angle = + 90.0;
        }
        else
        {
            diff_body_angle = - 90.0;
        }

        if ( final_body_pos.x <= -45.0 )
        {
            diff_body_angle *= -1.0;
        }
    }

    if ( rcsc::Body_TurnToAngle( (ball_pos - final_body_pos).th()
                                 + diff_body_angle ).execute( agent ) )
    {
        return true;
    }

    return true;
}


static
rcsc::Vector2D
s_get_position( rcsc::PlayerAgent * agent,
                const rcsc::Vector2D & ball_pos,
                double base_dist,
                const rcsc::Vector2D & self_pos,
                const rcsc::AbstractPlayerObject * target )
{
    return target->pos()
           + (ball_pos - target->pos()).setLengthVector( 2.0 )
           + rcsc::Vector2D( -2.0, 
                             rcsc::sign( target->pos().y ) * 2.0 );
}


#if 0
static
rcsc::Vector2D
s_get_position( rcsc::PlayerAgent * agent,
                const rcsc::Vector2D & ball_pos,
                double base_dist,
                const rcsc::Vector2D & self_pos )
{
    static const double MINIMUM_GOAL_X_DIST = 1.5;

    const double goal_half_width = rcsc::ServerParam::i().goalHalfWidth();

    const double alpha = std::atan2( goal_half_width, base_dist );

    const double min_x = (rcsc::ServerParam::i().ourTeamGoalLineX()
                          + MINIMUM_GOAL_X_DIST);

    if ( ball_pos.x <= rcsc::ServerParam::i().ourTeamGoalLineX()
         && ball_pos.absY() <= goal_half_width )
    {
        return rcsc::Vector2D( min_x, ball_pos.y );
    }

    rcsc::Vector2D goal_center = rcsc::ServerParam::i().ourTeamGoalPos();
    rcsc::Vector2D goal_left_post( goal_center.x, +goal_half_width );
    rcsc::Vector2D goal_right_post( goal_center.x, -goal_half_width );

    rcsc::Line2D line_1( ball_pos, goal_left_post );
    rcsc::Line2D line_2( ball_pos, goal_right_post );

    rcsc::AngleDeg line_dir = rcsc::AngleDeg::normalize_angle
                              ( rcsc::AngleDeg::normalize_angle
                                ( (goal_left_post - ball_pos).th().degree()
                                  + (goal_right_post - ball_pos).th().degree() )
                                / 2.0 );

    rcsc::Line2D line_m( ball_pos, line_dir );

    rcsc::Line2D goal_line( goal_left_post, goal_right_post );

    rcsc::Vector2D p = goal_line.intersection( line_m );

    if ( ! p.valid() )
    {
        if ( ball_pos.x > 0.0 )
        {
            return rcsc::Vector2D( min_x, goal_left_post.y );
        }
        else if ( ball_pos.x < 0.0 )
        {
            return rcsc::Vector2D( min_x, goal_right_post.y );
        }
        else
        {
            return rcsc::Vector2D( min_x, goal_center.y );
        }
    }

    // angle -> dist
    double dist_from_goal = (line_1.dist(p) + line_2.dist(p)) / 2.0
                            / std::sin(alpha);

    if ( dist_from_goal <= goal_half_width )
    {
        dist_from_goal = goal_half_width;
    }

    if ( (ball_pos - p).r() + 1.5 < dist_from_goal )
    {
        dist_from_goal = (ball_pos - p).r() + 1.5;
    }

    rcsc::AngleDeg current_position_line_dir = (self_pos - p).th();

    rcsc::AngleDeg position_error = line_dir - current_position_line_dir;

    const double danger_angle = 21.0;

    if ( position_error.abs() > danger_angle )
    {
        dist_from_goal *= (1.0 - (position_error.abs()
                                  - danger_angle)
                           / (180.0 - danger_angle) )
                          / 3.0;

        if ( dist_from_goal < goal_half_width - 1.0 )
        {
            dist_from_goal = goal_half_width - 1.0;
        }
    }

    rcsc::Vector2D result = p + (ball_pos - p).norm() * dist_from_goal;

    if ( result.x < min_x )
    {
        result.assign( min_x, result.y );
    }

    return result;
}
#endif
