// -*-c++-*-

/*!
  \file bhv_savior.h
  \brief aggressive goalie behavior
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bhv_savior.h"

#include <rcsc/geom/angle_deg.h>
#include <rcsc/common/server_param.h>
#include <rcsc/player/player_agent.h>
#include <rcsc/player/penalty_kick_state.h>
#if 0
#include <rcsc/action/body_intercept.h>
#else
#include "body_intercept_devel.h"
#endif
#include <rcsc/action/body_pass.h>
#include <rcsc/action/body_stop_ball.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_shoot.h>
#include <rcsc/action/body_hold_ball.h>
#include "body_go_to_point_devel.h"
#include <rcsc/action/body_turn_to_angle.h>
#include <rcsc/action/body_stop_dash.h>
#include <rcsc/action/body_dribble.h>
#include <rcsc/action/bhv_scan_field.h>
#include <rcsc/action/neck_turn_to_ball.h>
#include <rcsc-ext/action/neck_turn_to_ball_only.h>
#include <rcsc/action/neck_turn_to_point.h>
#include <rcsc/action/neck_turn_to_low_conf_teammate.h>
#include <rcsc/action/view_synch.h>

#include <rcsc-ext/action/bhv_find_player.h>
#include <rcsc-ext/action/bhv_find_ball.h>

#include <rcsc-ext/player/world_model_ext.h>
#include <rcsc/common/logger.h>
#include <rcsc/player/say_message_builder.h>
#include "bhv_goalie_basic_move.h"
#include "bhv_goalie_chase_ball.h"
#include "bhv_basic_tackle.h"
#include "body_clear_ball2008.h"
#include "body_get_ball.h"
#include "body_force_shoot.h"
#include "pass_course_table.h"
#include <algorithm>
#include <cmath>

#include "team_switch.h"

#if defined(USE_SMART_KICK)
# include "contrib/body_smart_kick.h"
#endif

#define USE_GET_BALL


bool
Bhv_Savior::execute( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

#if 1
    const bool use_conservative_mode = false;
#else
    const bool use_conservative_mode = true;
#endif

    const double our_team_goal_line_x = rcsc::ServerParam::i().ourTeamGoalLineX();
    switch( wm.gameMode().type() )
    {
    case rcsc::GameMode::PlayOn:
        return this->doPlayOnMove( agent, use_conservative_mode );

    case rcsc::GameMode::BeforeKickOff:
    case rcsc::GameMode::AfterGoal_:
        {
            const rcsc::Vector2D pos = rcsc::Vector2D( our_team_goal_line_x,
                                                       0.0 );

            if ( (pos - wm.self().pos()).r() > 1.0 )
            {
                agent->doMove( pos.x, pos.y );
            }
            else
            {
                rcsc::Bhv_ScanField().execute( agent );
            }
            agent->setNeckAction( new rcsc::Neck_TurnToBall );

            return true;
        }

    case rcsc::GameMode::PenaltyTaken_:
        {
            const rcsc::PenaltyKickState * pen = wm.penaltyKickState();

            if ( pen->currentTakerSide() == wm.ourSide() )
            {
                return rcsc::Bhv_ScanField().execute( agent );
            }
            else
            {
                return this->doPlayOnMove
                             ( agent, false,
                               true,
                               (pen->onfieldSide() != wm.ourSide()) );
            }
        }

    default:
        if ( wm.gameMode().type() == rcsc::GameMode::KickOff_
             && wm.gameMode().side() != wm.ourSide() )
        {
            return rcsc::Bhv_ScanField().execute( agent );
        }
        else
        {
            return this->doPlayOnMove( agent, use_conservative_mode );
        }
    }

    return false;
}

bool
Bhv_Savior::doPlayOnMove( rcsc::PlayerAgent * agent,
                          bool use_conservative_mode,
                          bool penalty_kick_mode,
                          bool reverse_x )
{
    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__ ": Bhv_Savior::execute().");

    const rcsc::WorldModel & wm = agent->world();

    agent->setNeckAction( new rcsc_ext::Neck_TurnToBallOnly );

    //
    // if catchable, do catch
    //
    const double max_ball_error = 0.12;

    bool ball_cachable_if_rearch
        = (wm.self().goalie()
           && wm.time().cycle()
            >= agent->effector().getCatchTime().cycle()
               + rcsc::ServerParam::i().catchBanCycle()
           && wm.ball().pos().x < rcsc::ServerParam::i().ourPenaltyAreaLineX()
                                   + rcsc::ServerParam::i().ballSize() * 2
                                   - max_ball_error
           && wm.ball().pos().absY()
              < rcsc::ServerParam::i().penaltyAreaHalfWidth()
                + rcsc::ServerParam::i().ballSize() * 2
                - max_ball_error
           && wm.ball().pos().x
              >= rcsc::ServerParam::i().ourTeamGoalLineX() - 1.0e-10);


    // if catchable situation
    if ( ball_cachable_if_rearch
         && wm.ball().distFromSelf() < wm.self().catchableArea() - 0.01 )
    {
        // if near opponent is not exist, catch the ball,
        // otherwise don't catch it.
        //
        // (this check is for avoiding back pass foul)
        if ( penalty_kick_mode
             || (wm.ball().pos().x <= 40.0
                 || ! wm.getPlayerCont
                      ( new rcsc::AndPlayerPredicate
                        ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ),
                          new rcsc::PointNearPlayerPredicate( wm.self().pos(),
                                                              5.0 ) ) )
                      .empty()) )
        {
            return agent->doCatch();
        }
    }


    //
    // if kickable, do kick
    //
    if ( wm.self().isKickable() )
    {
        return this->doKick( agent );
    }


    //
    // tackle
    //
    if ( Bhv_BasicTackle( 0.85, 160.0 ).execute( agent ) )
    {
        return true;
    }


    //
    // chase ball
    //
    if ( Bhv_GoalieChaseBall::is_ball_chase_situation( agent ) )
    {
#if defined(USE_GET_BALL)
        return Body_GetBall().execute( agent );
#else
# if 0
        return Bhv_GoalieChaseBall().execute( agent );
# else
        return rcsc::Body_Intercept().execute( agent );
# endif
#endif
    }
#if defined(USE_GET_BALL)
    else if ( Body_GetBall::gettable_step( wm )
#else
    else if ( wm.interceptTable()->selfReachCycle()
#endif
              <= wm.interceptTable()->teammateReachCycle()
              && wm.interceptTable()->selfReachCycle() + 1
              < wm.interceptTable()->opponentReachCycle()
              && wm.gameMode().type() == rcsc::GameMode::PlayOn )
    {
#if defined(USE_GET_BALL)
        return Body_GetBall().execute( agent );
#else
# if 0
        return Bhv_GoalieChaseBall().execute( agent );
# else
        return rcsc::Body_Intercept().execute( agent );
# endif
#endif
    }


    //
    // Check Ball
    //
    long ball_premise_accuracy = 2;

    if ( wm.self().pos().x >= -30.0 )
    {
        ball_premise_accuracy = 6;
    }

    if ( wm.ball().seenPosCount() > ball_premise_accuracy )
    {
        return rcsc_ext::Bhv_FindBall().execute( agent );
    }


    const int teammate_rearch_cycle = wm.interceptTable()->teammateReachCycle();
    const int opponent_rearch_cycle = wm.interceptTable()->opponentReachCycle();
#if defined(USE_GET_BALL)
    const int self_rearch_cycle = Body_GetBall::gettable_step( wm );
#else
    const int self_rearch_cycle = wm.interceptTable()->selfReachCycle();
#endif

    const int predict_step = std::min( opponent_rearch_cycle,
                                       std::min( teammate_rearch_cycle,
                                                 self_rearch_cycle ) );

    const rcsc::Vector2D ball_pos = wm.ball().inertiaPoint( predict_step );

    bool have_advantage = (std::min( teammate_rearch_cycle, self_rearch_cycle )
                           < opponent_rearch_cycle);


    //
    // conservative mode
    //
    if ( use_conservative_mode
         && ((have_advantage && wm.defenseLineX() <= - 30.0)
             || (! have_advantage && wm.defenseLineX() <= -20.0)) )
    {
        return Bhv_GoalieBasicMove().execute( agent );
    }


    //
    // positioning
    //
    bool in_region;
    bool dangerous;
    bool aggressive;
    bool emergent_advance;

    rcsc::Vector2D best_position = this->getBasicPosition
                                           ( agent,
                                             use_conservative_mode,
                                             penalty_kick_mode,
                                             reverse_x,
                                             &in_region,
                                             &dangerous,
                                             &aggressive,
                                             &emergent_advance );

#if 0
    //
    // attack to near ball
    //
    if ( ! dangerous
         && ball_cachable_if_rearch
         && wm.ball().distFromSelf() < 5.0 )
    {
# if defined(USE_GET_BALL)
        return Body_GetBall().execute( agent );
# elif 0
        return Bhv_GoalieChaseBall().execute( agent );
# else
        return Body_Intercept().execute( agent );
# endif
    }
#endif


    agent->debugClient().addLine( wm.self().pos(), best_position );


    double max_position_error = 0.70;

    double dash_power = rcsc::ServerParam::i().maxPower();
#if 0
    bool back_dash = true;
#endif

    if ( aggressive )
    {
        if ( (ball_pos - wm.self().pos()).r() >= 30.0 )
        {
            max_position_error = 2.0;
        }
    }


    if ( emergent_advance )
    {
        double diff = (best_position - wm.self().pos()).r();

        if ( diff >= 10.0 )
        {
            // for speedy movement
            max_position_error = 1.9;
        }
        else if ( diff >= 5.0 )
        {
            // for speedy movement
            max_position_error = 1.0;
        }
    }
    else if ( dangerous )
    {
        double diff = (best_position - wm.self().pos()).r();

        if ( diff >= 10.0 )
        {
            // for speedy movement
            max_position_error = 1.9;
        }
        else if ( diff >= 5.0 )
        {
            // for speedy movement
            max_position_error = 1.0;
        }
    }

    if ( Body_GoToPoint_Devel( best_position,
                               max_position_error,
                               dash_power ).execute( agent ) )
    {
        return true;
    }


    //
    // turn body angle against ball
    //
    const rcsc::Vector2D final_body_pos = wm.self().inertiaFinalPoint();

    rcsc::AngleDeg diff_body_angle = 0.0;

    if ( (ball_pos - final_body_pos).r() <= 10.0
         && (wm.existKickableOpponent() || wm.existKickableTeammate()) )
    {
        if ( final_body_pos.y > 0.0 )
        {
            diff_body_angle = + 90.0;
        }
        else
        {
            diff_body_angle = - 90.0;
        }

        if ( translateXIsLessEqualsTo( reverse_x, final_body_pos.x, -45.0 ) )
        {
            diff_body_angle *= -1.0;
        }
    }

    if ( rcsc::Body_TurnToAngle( (ball_pos - final_body_pos).th()
                                 + diff_body_angle ).execute( agent ) )
    {
        return true;
    }


#if 0
    //
    // stop
    //
    if ( rcsc::Body_StopDash( true ).execute( agent ) )
    {
        return true;
    }
#endif


    //
    // go to position with minimum error
    //
    if ( Body_GoToPoint_Devel( best_position,
                               0.2,
                               dash_power ).execute( agent )
         || Body_GoToPoint_Devel( best_position,
                                  0.1,
                                  dash_power ).execute( agent ) )
    {
        return true;
    }

#if 1
    //
    // stop
    //
    if ( rcsc::Body_StopDash( true ).execute( agent ) )
    {
        return true;
    }
#endif

    return true;
}


bool
Bhv_Savior::doKick( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    if ( wm.existKickableOpponent() )
    {
        Body_ClearBall2008().execute( agent );
        agent-> setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

        agent->debugClient().addMessage( "clear ball" );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "opponent kickable, clear ball" );
        return true;
    }


#if 1
    const PassCourseTable & table = PassCourseTable::instance
                                    ( wm, ! wm.teammates().empty() );

    PassCourseTable::CourseSelection result = table.get_first_selection( agent );

    const rcsc::Vector2D goal_pos = rcsc::ServerParam::i().theirTeamGoalPos();
    agent-> setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

    switch( result.type )
    {
    case PassCourseTable::CourseSelection::Shoot:
        if ( Body_ForceShoot().execute( agent ) )
        {
            return true;
        }

        break;

    case PassCourseTable::CourseSelection::Push:
      {
        rcsc::Vector2D dribble_target;

        if ( wm.self().pos().x > + 35.0 )
        {
            dribble_target = rcsc_ext::ServerParam_opponentTeamFarGoalPostPos
                                                          ( wm.self().pos() );
        }
        else
        {
            dribble_target = wm.self().pos() + rcsc::Vector2D( 10.0, 0.0 );
        }

        rcsc::Body_Dribble( dribble_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower(),
                            3 ).execute( agent );
        return true;
        break;
      }

    case PassCourseTable::CourseSelection::FindPassReceiver:
    case PassCourseTable::CourseSelection::Pass:
        {
            const rcsc::AbstractPlayerObject * target;
            target = wm.teammate( result.player_number );

            if ( rcsc_ext::getDistPlayerNearestToPoint
                           ( wm, wm.self().pos(), wm.opponents() ) > 4.0 )
            {
                // XXX: if ball kickable at next cycle, no need to stop ball
                // XXX: Neck_FindPlayer(target)
                if ( rcsc::Body_StopBall().execute( agent )
                     || rcsc_ext::Bhv_FindPlayer( target ).execute( agent ) )
                {
                    agent->debugClient().addMessage( "find" );
                    agent->debugClient().setTarget( target->unum() );
                }
            }

            table.setDebugPassRoute( agent );

            const rcsc::Vector2D target_point = target->pos();
#if defined(USE_SMART_KICK)
            rcsc::Body_SmartKick( target_point,
                                  rcsc::ServerParam::i().maxPower(),
                                  rcsc::ServerParam::i().maxPower() * 0.96,
                                  3 ).execute( agent );
#else
            rcsc::Body_KickMultiStep( target_point,
                                      2.5,
                                      false ).execute( agent );
#endif

            if ( agent->config().useCommunication()
                 && result.player_number != rcsc::Unum_Unknown )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    __FILE__
                                    ": execute() set pass communication." );
            }

            rcsc::Vector2D target_buf = target_point - agent->world().self().pos();
            target_buf.setLength( 1.0 );

            agent->addSayMessage( new rcsc::PassMessage( result.player_number,
                                                         target_point + target_buf,
                                                         agent->effector().queuedNextBallPos(),
                                                         agent->effector().queuedNextBallVel() ) );
            return true;
            break;
        }

    case PassCourseTable::CourseSelection::Hold:
        Body_ClearBall2008().execute( agent );
        agent-> setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

        agent->debugClient().addMessage( "clear ball" );
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            "clear ball" );
        return true;
        break;
    }
#else
    const PassCourseTable & table = PassCourseTable::instance
                                    ( wm, ! wm.teammates().empty() );

    const PassCourseTable::CourseSelection first_pass = table.get_first_selection( agent );

    const rcsc::Vector2D goal_pos = rcsc::ServerParam::i().theirTeamGoalPos();
    agent-> setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

    switch( first_pass.type )
    {
    case PassCourseTable::CourseSelection::Shoot:
        if ( Body_ForceShoot().execute( agent ) )
        {
            return true;
        }

        break;

    case PassCourseTable::CourseSelection::Push:
      {
        rcsc::Vector2D dribble_target;

        if ( wm.self().pos().x > + 35.0 )
        {
            dribble_target = rcsc_ext::ServerParam_opponentTeamFarGoalPostPos
                                                          ( wm.self().pos() );
        }
        else
        {
            dribble_target = wm.self().pos() + rcsc::Vector2D( 10.0, 0.0 );
        }

        rcsc::Body_Dribble( dribble_target,
                            1.0,
                            rcsc::ServerParam::i().maxPower(),
                            3 ).execute( agent );
        return true;
        break;
      }

    case PassCourseTable::CourseSelection::FindPassReceiver:
    case PassCourseTable::CourseSelection::Pass:
        {
            const rcsc::AbstractPlayerObject * target_player;
            target_player = wm.teammate( first_pass.player_number );

            if ( rcsc_ext::getDistPlayerNearestToPoint
                           ( wm, wm.self().pos(), wm.opponents() ) > 4.0
                && first_pass.type == PassCourseTable::CourseSelection::FindPassReceiver )
            {
                if ( ! agent->effector().queuedNextBallKickable()
                     && wm.ball().vel().r() >= 0.1 )
                {
                    if ( rcsc::Body_StopBall().execute( agent ) )
                    {
                        agent-> setNeckAction( new rcsc::Neck_TurnToPoint
                                               ( target_player->pos() ) );

                        agent->debugClient().addMessage( "stop ball" );
                        agent->debugClient().setTarget( target_player->unum() );

                        rcsc::dlog.addText( rcsc::Logger::TEAM,
                                            "stop ball" );

                        return true;
                    }
                }

                if ( rcsc_ext::Bhv_FindPlayer( target_player )
                     .execute( agent ) )
                {
                    agent-> setNeckAction( new rcsc::Neck_TurnToPoint
                                           ( target_player->pos() ) );

                    agent->debugClient().addMessage( "find" );
                    agent->debugClient().setTarget( target_player->unum() );

                    rcsc::dlog.addText( rcsc::Logger::TEAM,
                                        "find" );

                    return true;
                }
            }

            table.setDebugPassRoute( agent );

            const rcsc::Vector2D target_point = target_player->pos();
            rcsc::Body_KickMultiStep( target_point,
                                      2.5,
                                      false ).execute( agent );

            if ( agent->config().useCommunication()
                 && first_pass.player_number != rcsc::Unum_Unknown )
            {
                rcsc::dlog.addText( rcsc::Logger::TEAM,
                                    __FILE__
                                    ": execute() set pass communication." );
            }

            rcsc::Vector2D target_buf = target_point - agent->world().self().pos();
            target_buf.setLength( 1.0 );

            agent->addSayMessage( new rcsc::PassMessage( first_pass.player_number,
                                                         target_point + target_buf,
                                                         agent->effector().queuedNextBallPos(),
                                                         agent->effector().queuedNextBallVel() ) );
            return true;
            break;
        }

    case PassCourseTable::CourseSelection::Hold:
        Body_ClearBall2008().execute( agent );
        agent-> setNeckAction( new rcsc::Neck_TurnToLowConfTeammate() );

        agent->debugClient().addMessage( "clear ball" );
        rcsc::dlog.addText( rcsc::Logger::TEAM, "clear ball" );
        return true;
        break;
    }
#endif

    return false;
}

/*!
  \class ContainsPlayerPredicate
  \brief check if target player is in region
*/
template<typename T>
class ContainsPlayerPredicate
    : public rcsc::PlayerPredicate
{
private:
    //! geomotry to check
    T M_geom;

public:
    /*!
      \brief construct with the geometry object
      \param geometry object for checking
     */
    ContainsPlayerPredicate( const T & geom )
        : M_geom( geom )
      {
      }

    /*!
      \brief predicate function
      \param p const reference to the target player object
      \return true if target player is in the region
     */
    bool operator()( const rcsc::AbstractPlayerObject & p ) const
      {
          return M_geom.contains( p.pos() );
      }
};

rcsc::Vector2D
Bhv_Savior::getBasicPosition( rcsc::PlayerAgent * agent,
                              bool use_conservative_mode,
                              bool penalty_kick_mode,
                              bool reverse_x,
                              bool * in_region,
                              bool * dangerous,
                              bool * aggressive,
                              bool * emergent_advance ) const
{
    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    static const double AGGRESSIVE_POSITIONING_STAMINA_RATE_THRESHOLD = 0.85;

    const int teammate_rearch_cycle = wm.interceptTable()->teammateReachCycle();
    const int opponent_rearch_cycle = wm.interceptTable()->opponentReachCycle();
#if defined(USE_GET_BALL)
    const int self_rearch_cycle = Body_GetBall::gettable_step( wm );
#else
    const int self_rearch_cycle = wm.interceptTable()->selfReachCycle();
#endif

    const int predict_step = std::min( opponent_rearch_cycle,
                                       std::min( teammate_rearch_cycle,
                                                 self_rearch_cycle ) );

    const rcsc::Vector2D ball_pos = wm.ball().inertiaPoint( predict_step );
    const rcsc::Vector2D self_pos = wm.self().pos();

    rcsc::Vector2D positioning_base_pos = ball_pos;

#if 0
    const rcsc::AbstractPlayerObject * holder = wm.interceptTable()->fastestOpponent();

    if ( holder )
    {
        double ball_rate = 3.0 / 4.0;

        positioning_base_pos = ball_pos * ball_rate
                               + holder->pos() * (1.0 - ball_rate);
    }
#endif


    rcsc::Rect2D penalty_area;
    rcsc::Rect2D conservative_area;

    if ( reverse_x )
    {
        penalty_area.assign( param.theirPenaltyAreaLineX(),
                             - param.penaltyAreaHalfWidth(),
                             param.penaltyAreaLength(),
                             param.penaltyAreaWidth() );

        conservative_area.assign( param.theirPenaltyAreaLineX() - 15.0,
                                  - param.penaltyAreaHalfWidth(),
                                  param.penaltyAreaLength() + 15.0,
                                  param.penaltyAreaWidth() );
    }
    else
    {
        penalty_area.assign( param.ourTeamGoalLineX(),
                             - param.penaltyAreaHalfWidth(),
                             param.penaltyAreaLength(),
                             param.penaltyAreaWidth() );

        conservative_area.assign( param.ourTeamGoalLineX(),
                                  - param.penaltyAreaHalfWidth(),
                                  param.penaltyAreaLength() + 15.0,
                                  param.penaltyAreaWidth() );
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__
                        ": number of teammates in conservative area is %d",
                        wm.getPlayerCont( new rcsc::AndPlayerPredicate
                                                                                                                   ( new rcsc::TeammatePlayerPredicate( wm ),
                                                                                                                     new ContainsPlayerPredicate<rcsc::Rect2D>
                                                                                                                     ( conservative_area ) ) ).size() );

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__
                        ": ball in area = %d",
                        (int)(conservative_area.contains( wm.ball().pos() )) );

#if 0
    double base_dist = 22.0;
#else
    double base_dist = 14.0;
#endif

#if 1
    if ( penalty_area.contains( wm.ball().pos() )
         && wm.getPlayerCont
               ( new rcsc::AndPlayerPredicate
                 ( new rcsc::TeammatePlayerPredicate( wm ),
                   new ContainsPlayerPredicate<rcsc::Rect2D>
                   ( penalty_area ) ) ).size() >= 3 )
    {
        base_dist = std::min( base_dist,
                              std::max( wm.defenseLineX()
                                        - param.ourPenaltyAreaLineX() - 2.0,
                                        2.0 ) );
    }
#endif

#if 1
    if ( ! penalty_area.contains( wm.ball().pos() )
         && wm.getPlayerCont
               ( new rcsc::AndPlayerPredicate
                 ( new rcsc::TeammatePlayerPredicate( wm ),
                   new ContainsPlayerPredicate<rcsc::Rect2D>
                   ( conservative_area ) ) ).size() >= 3 )
    {
        base_dist = 5.0;
    }
#endif

    // aggressive positioning
    const double additional_forward_positioning_max = 18.0;

    bool emergent_advance_mode = false;
#if 1
    const rcsc::AbstractPlayerObject * fastestOpponent = wm.interceptTable()->fastestOpponent();

    if ( ball_pos.x < 0.0
         && ball_pos.x >= param.ourPenaltyAreaLineX()
         && wm.defenseLineX() <= ball_pos.x
         && ball_pos.x <= wm.defenseLineX() + 5.0
#if 1
         && opponent_rearch_cycle <= teammate_rearch_cycle - 2
#else
         && opponent_rearch_cycle <= teammate_rearch_cycle - 1
#endif
         && fastestOpponent
         && wm.getPlayerCont
               ( new rcsc::AndPlayerPredicate
                 ( new rcsc::TeammatePlayerPredicate( wm ),
                   new rcsc::XCoordinateBackwardPlayerPredicate
                   ( fastestOpponent->pos().x ) ,
#if 0
                   new rcsc::YCoordinatePlusPlayerPredicate
                   ( fastestOpponent->pos().y - 1.0 ),
                   new rcsc::YCoordinateMinusPlayerPredicate
                   ( fastestOpponent->pos().y + 1.0 ) ) ).empty() )
#else
                   new rcsc::YCoordinatePlusPlayerPredicate
                   ( fastestOpponent->pos().y - 2.0 ),
                   new rcsc::YCoordinateMinusPlayerPredicate
                   ( fastestOpponent->pos().y + 2.0 ) ) ).empty() )
#endif
    {
        emergent_advance_mode = true;

        if ( emergent_advance )
        {
            *emergent_advance = true;
        }

        agent->debugClient().addMessage( "EmergentAdvance" );
    }
#endif

    if ( penalty_kick_mode )
    {
        base_dist = (rcsc::ServerParam::i().ourTeamGoalPos() - ball_pos).r()
                    - 3.0;
    }
    else if ( emergent_advance_mode )
    {
        base_dist = (rcsc::ServerParam::i().ourTeamGoalPos() - ball_pos).r()
                    - 7.0;
    }
    else if ( wm.defenseLineX() >= - additional_forward_positioning_max
              && wm.self().stamina()
                 >= rcsc::ServerParam::i().staminaMax()
                             * AGGRESSIVE_POSITIONING_STAMINA_RATE_THRESHOLD )
    {
        base_dist += std::min( wm.defenseLineX(), 0.0 )
                     + additional_forward_positioning_max;

        if ( aggressive )
        {
            *aggressive = true;
        }
    }
    else
    {
        if ( aggressive )
        {
            *aggressive = false;
        }
    }

    rcsc::dlog.addText( rcsc::Logger::TEAM,
                        __FILE__ ": base dist = %f", base_dist );

    if ( in_region )
    {
        *in_region = true;
    }

    if ( dangerous )
    {
        *dangerous = false;
    }

    if ( ! reverse_x )
    {
        return getBasicPositionFromBall( agent,
                                         positioning_base_pos,
                                         base_dist,
                                         wm.self().pos(),
                                         wm.defenseLineX(),
                                         use_conservative_mode,
                                         penalty_kick_mode,
                                         in_region,
                                         dangerous );
    }
    else
    {
        rcsc::Vector2D reversed_ball_pos( - positioning_base_pos.x,
                                          positioning_base_pos.y );

        rcsc::Vector2D self_pos( - wm.self().pos().x, wm.self().pos().y );

        rcsc::Vector2D result_pos = getBasicPositionFromBall
                                    ( agent,
                                      reversed_ball_pos,
                                      base_dist,
                                      self_pos,
                                      translateX( reverse_x,
                                                  wm.defenseLineX() ),
                                      use_conservative_mode,
                                      penalty_kick_mode,
                                      in_region,
                                      dangerous );

        return rcsc::Vector2D( - result_pos.x, result_pos.y );
    }
}

rcsc::Vector2D
Bhv_Savior::getBasicPositionFromBall( rcsc::PlayerAgent * agent,
                                      const rcsc::Vector2D & ball_pos,
                                      double base_dist,
                                      const rcsc::Vector2D & self_pos,
                                      double defense_line_x,
                                      bool use_conservative_mode,
                                      bool no_danger_angle,
                                      bool * in_region,
                                      bool * dangerous ) const
{
    static const double MINIMUM_GOAL_X_DIST = 1.5;

    const double our_team_goal_line_x = rcsc::ServerParam::i().ourTeamGoalLineX();
    const double goal_half_width = rcsc::ServerParam::i().goalHalfWidth();

    const double alpha = std::atan2( goal_half_width, base_dist );

    const double min_x = (rcsc::ServerParam::i().ourTeamGoalLineX()
                          + MINIMUM_GOAL_X_DIST);

    if ( ball_pos.x <= rcsc::ServerParam::i().ourTeamGoalLineX()
         && ball_pos.absY() <= goal_half_width )
    {
        return rcsc::Vector2D( min_x, ball_pos.y );
    }

    rcsc::Vector2D goal_center = rcsc::ServerParam::i().ourTeamGoalPos();
    rcsc::Vector2D goal_left_post( goal_center.x, +goal_half_width );
    rcsc::Vector2D goal_right_post( goal_center.x, -goal_half_width );

    rcsc::Line2D line_1( ball_pos, goal_left_post );
    rcsc::Line2D line_2( ball_pos, goal_right_post );

    rcsc::AngleDeg line_dir = rcsc::AngleDeg::normalize_angle
                              ( rcsc::AngleDeg::normalize_angle
                                ( (goal_left_post - ball_pos).th().degree()
                                  + (goal_right_post - ball_pos).th().degree() )
                                / 2.0 );

    rcsc::Line2D line_m( ball_pos, line_dir );

    rcsc::Line2D goal_line( goal_left_post, goal_right_post );

    rcsc::Vector2D p = goal_line.intersection( line_m );

    if ( ! p.valid() )
    {
        if ( ball_pos.x > 0.0 )
        {
            return rcsc::Vector2D( min_x, goal_left_post.y );
        }
        else if ( ball_pos.x < 0.0 )
        {
            return rcsc::Vector2D( min_x, goal_right_post.y );
        }
        else
        {
            return rcsc::Vector2D( min_x, goal_center.y );
        }
    }

    if ( in_region != static_cast<bool *>(0) )
    {
        *in_region = false;
    }

    // angle -> dist
    double dist_from_goal = (line_1.dist(p) + line_2.dist(p)) / 2.0
                            / std::sin(alpha);

    if ( dist_from_goal <= goal_half_width )
    {
        dist_from_goal = goal_half_width;
    }


    if ( use_conservative_mode
         && our_team_goal_line_x + dist_from_goal > defense_line_x - 10.0 )
    {
        dist_from_goal = defense_line_x - our_team_goal_line_x - 10.0;
    }


    if ( (ball_pos - p).r() + 1.5 < dist_from_goal )
    {
        dist_from_goal = (ball_pos - p).r() + 1.5;
    }

    rcsc::AngleDeg current_position_line_dir = (self_pos - p).th();

    rcsc::AngleDeg position_error = line_dir - current_position_line_dir;

    const double danger_angle = 21.0;

    if ( ! no_danger_angle
         && position_error.abs() > danger_angle )
    {
        dist_from_goal *= (1.0 - (position_error.abs()
                                  - danger_angle)
                           / (180.0 - danger_angle) )
                          / 3.0;

        if ( dist_from_goal < goal_half_width - 1.0 )
        {
            dist_from_goal = goal_half_width - 1.0;
        }

        if ( dangerous != static_cast<bool *>(0) )
        {
            *dangerous = true;
        }
    }

    rcsc::Vector2D result = p + (ball_pos - p).norm() * dist_from_goal;

    if ( result.x < min_x )
    {
        result.assign( min_x, result.y );
    }

    return result;
}

double
Bhv_Savior::translateX( bool reverse, double x ) const
{
    if ( reverse )
    {
        return -x;
    }
    else
    {
        return x;
    }
}

bool
Bhv_Savior::translateXIsLessEqualsTo( bool reverse,
                                      double x, double threshold ) const
{
    if ( reverse )
    {
        return (-x >= threshold);
    }
    else
    {
        return (x <= threshold);
    }
}
