// -*-c++-*-

/*!
  \file body_clear_ball2008.cpp
  \brief kick the ball to escape from dangerous situation
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_clear_ball2008.h"

#include "simulator.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/math_util.h>

#include "rcsc-ext/player/player_evaluator.h"
#include "rcsc-ext/player/world_model_ext.h"

#include <rcsc/action/body_kick_one_step.h>
#include <rcsc/action/body_kick_two_step.h>

#include <boost/shared_ptr.hpp>

#include <cmath>

static
rcsc::AngleDeg
getClearCourse( const rcsc::WorldModel & wm,
                double safe_angle,
                int recursive_count );

static
double
getFreeAngle( const rcsc::WorldModel & wm, const rcsc::AngleDeg & angle );

/*-------------------------------------------------------------------*/
/*!

*/
Body_ClearBall2008::Body_ClearBall2008()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Body_ClearBall2008::execute( rcsc::PlayerAgent * agent )
{
    // TODO:
    // . if result free cycles will be same steps, choice by free angle width
    // . check tackle

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();


    // enforce kick
    if ( wm.existKickableOpponent() )
    {
        return rcsc::Body_KickOneStep
               ( wm.ball().pos() + rcsc::Vector2D( 10.0, 0.0 ),
                 param.ballSpeedMax() ).execute( agent );
    }

    rcsc::AngleDeg clear_angle = getClearCourse( wm,
                                                 15.0, /* safe angle */
                                                 3 /* recursive count */ );

    const rcsc::Vector2D kick_target = wm.ball().pos()
			               + rcsc::Vector2D::polar2vector
                                         ( 30.0, clear_angle );

    agent->debugClient().addLine( wm.ball().pos(), kick_target );

    if ( rcsc_ext::AbstractPlayerCont_getMinimumEvaluation
         ( wm.getPlayerCont
           ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ) ),
           new rcsc_ext::DistFromPosPlayerEvaluator( wm.ball().pos() ) )
         < 3.0 )
    {
        return rcsc::Body_KickOneStep( kick_target,
                                       param.ballSpeedMax() ).execute( agent );
    }
    else
    {
        return rcsc::Body_KickTwoStep( kick_target,
                                       param.ballSpeedMax() ).execute( agent );
    }
}

static
double
getFreeAngle( const rcsc::WorldModel & wm, const rcsc::AngleDeg & angle )
{
    return std::min( rcsc_ext::AbstractPlayerCont_getMinimumEvaluation
                     ( wm.getPlayerCont
                       ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ) ),
                       new rcsc_ext::AbsAngleDiffPlayerEvaluator
                       ( wm.ball().pos(), angle ) ),
                     360.0 );
}


static
rcsc::AngleDeg
getClearCourse( const rcsc::WorldModel & wm,
                double safe_angle,
                int recursive_count )

{
    const rcsc::ServerParam & param = rcsc::ServerParam::i();

    double sign = rcsc::sign( wm.ball().pos().y );

    rcsc::Vector2D clear_point;
    rcsc::AngleDeg clear_dir;

    // check clear point 1
    clear_point.assign( wm.self().pos().x + 50.0,
                        sign * (param.pitchHalfWidth() - 2.0) );

    if ( (wm.self().pos() - clear_point).r() >= 10.0 )
    {
        clear_dir = (clear_point - wm.ball().pos()).th();

        if ( getFreeAngle( wm, clear_dir ) >= safe_angle )
        {
            return clear_dir;
        }
    }

    // check clear point 2
    clear_point.assign( wm.self().pos().x + 40.0,
                        sign * (param.pitchHalfWidth() - 2.0) );

    if ( (wm.self().pos() - clear_point).r() >= 10.0 )
    {
        clear_dir = (clear_point - wm.ball().pos()).th();

        if ( getFreeAngle( wm, clear_dir ) >= safe_angle )
        {
            return clear_dir;
        }
    }

    // check clear point 3
    clear_point.assign( wm.self().pos().x + 30.0,
                        sign * (param.pitchHalfWidth() - 2.0) );

    if ( (wm.self().pos() - clear_point).r() >= 10.0 )
    {
        clear_dir = (clear_point - wm.ball().pos()).th();

        if ( getFreeAngle( wm, clear_dir ) >= safe_angle )
        {
            return clear_dir;
        }
    }

    // check clear point 4
    clear_point.assign( wm.self().pos().x + 20.0,
                        sign * (param.pitchHalfWidth() - 2.0) );

    if ( (wm.self().pos() - clear_point).r() >= 10.0 )
    {
        clear_dir = (clear_point - wm.ball().pos()).th();

        if ( getFreeAngle( wm, clear_dir ) >= safe_angle )
        {
            return clear_dir;
        }
    }

    // search from 30 degree to 75 degree
    for ( double dir = 30.0; dir <= 75.0 + rcsc::EPS; dir += 1.0 )
    {
        if ( getFreeAngle( wm, sign * dir ) >= safe_angle )
        {
            return sign * dir;
        }
    }

    // search from 30 degree to -30 degree
    for ( double dir = 30.0; dir >= -30.0 - rcsc::EPS; dir -= 1.0 )
    {
        if ( getFreeAngle( wm, sign * dir ) >= safe_angle )
        {
            return sign * dir;
        }
    }

    if ( recursive_count > 0 )
    {
        rcsc::dlog.addText( rcsc::Logger::TEAM,
                            __FILE__ " recursive: %d", recursive_count );

        return getClearCourse( wm,
                               safe_angle * 0.7,
                               recursive_count - 1 );
    }
    else
    {
        rcsc::AngleDeg goal_away = (wm.ball().pos()
                                    - param.ourTeamGoalPos()).th();

        if ( std::fabs( wm.ball().pos().y + goal_away.sin() * 30.0 )
             < param.pitchHalfWidth() )
        {
            // goal away
            rcsc::dlog.addText( rcsc::Logger::TEAM, __FILE__ ": goal_away" );
            return goal_away;
        }
        else
        {
            // beside line
            rcsc::dlog.addText( rcsc::Logger::TEAM, __FILE__ ": beside line" );

            if ( wm.self().pos().absY() <= 25.0 )
            {
                return + rcsc::sign( wm.ball().pos().y ) * 3.0;
            }
            else
            {
                return - rcsc::sign( wm.ball().pos().y ) * 3.0;
            }
        }
    }

    return 0.0;
}
