// -*-c++-*-

/*!
  \file body_shoot_test.cpp
  \brief kick ball to opponent's goal
*/

/*
 *Copyright:

 Copyright (C) Hiroki SHIMORA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "body_shoot_test.h"

#include <rcsc/player/player_agent.h>
#include <rcsc/player/player_predicate.h>
#include <rcsc/common/server_param.h>

#include <rcsc/common/logger.h>
#include <rcsc/math_util.h>

#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/body_kick_one_step.h>

#include "simulator_util.h"

#include <boost/shared_ptr.hpp>
#include <utility>
#include <algorithm>

static
std::pair< bool, double >
evaluateSituation( const SimulatorUtil::SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent );

/*-------------------------------------------------------------------*/
/*!

*/
Body_Shoot_Test::Body_Shoot_Test()
{
}

/*-------------------------------------------------------------------*/
/*!

*/
bool
Body_Shoot_Test::execute( rcsc::PlayerAgent * agent )
{
    rcsc::dlog.addText( rcsc::Logger::ACTION,
                        __FILE__ ": execute()" );

    const rcsc::WorldModel & wm = agent->world();
    const rcsc::ServerParam & param = rcsc::ServerParam::i();


    //
    // create candidates
    //
    SimulatorUtil::KickPlanCont kick_candidates;

    double goal_half_width = param.goalHalfWidth() - 0.5;

    const int N = 10;
    for ( int i = 0; i <= N; ++i )
    {
        const double y = - goal_half_width + i * (goal_half_width * 2.0 / N);

        const rcsc::Vector2D target_point( param.theirTeamGoalLineX(), y );

        double target_ball_speed = param.ballSpeedMax();

        const rcsc::Vector2D result_pos = wm.self().pos();
        const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                          ( target_ball_speed,
                                            (target_point
                                             - wm.self().pos()).th() );
                                           
        const int kick_step = 2;

        kick_candidates.push_back( SimulatorUtil::KickPlan
                                   ( boost::shared_ptr< rcsc::BodyAction >
                                     ( new rcsc::Body_KickMultiStep
                                       ( target_point,
                                         target_ball_speed,
                                         false ) ),
                                     result_pos,
                                     result_vel,
                                     kick_step ) );
    }


    //
    // one step kick
    //
    for ( int i = 0; i <= N; ++i )
    {
        const double y = - goal_half_width + i * (goal_half_width * 2.0 / N);

        const rcsc::Vector2D target_point( param.theirTeamGoalLineX(), y );

        double target_ball_speed = rcsc::Body_KickOneStep::get_max_possible_vel
                                   ( (target_point - wm.ball().pos()).th(),
                                     wm.self().kickRate(),
                                     wm.ball().vel() ).r();

        boost::shared_ptr< rcsc::BodyAction >
            kick( new rcsc::Body_KickOneStep( target_point,
                                              target_ball_speed ) );

        const rcsc::Vector2D result_pos = wm.ball().pos();
        const rcsc::Vector2D result_vel = rcsc::Vector2D::polar2vector
                                          ( target_ball_speed,
                                            (target_point
                                             - wm.ball().pos()).th() );
                                           
        const int kick_step = 1;

        kick_candidates.push_back( SimulatorUtil::KickPlan
                                   ( kick,
                                     result_pos,
                                     result_vel,
                                     kick_step ) );
    }


    //
    // get the best plan
    //
    boost::shared_ptr< SimulatorUtil::SimulatorEstimatedSituation > best;

    best = SimulatorUtil::getBestChoice
                          ( evaluateSituation,
                            kick_candidates,
                            wm.getPlayerCont
                            ( new rcsc::OpponentOrUnknownPlayerPredicate( wm ) ),
                            agent,
                            true, 0,
                            true, 0,
                            false, 0 );


    //
    // execute
    //
    if ( best )
    {
        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": shoot route found" );

        return best->plan.getAction()->execute( agent );
    }
    else
    {
        rcsc::dlog.addText( rcsc::Logger::ACTION,
                            __FILE__ ": shoot route not found" );

        return false;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
static
std::pair< bool, double >
evaluateSituation( const SimulatorUtil::SimulatorEstimatedSituation & s,
                   rcsc::PlayerAgent * agent )
{
    const rcsc::ServerParam & param = rcsc::ServerParam::i();
    const rcsc::WorldModel & wm = agent->world();

    long opp_team_access_step = std::min( s.opponent_access_step,
                                          s.unknown_access_step );

    rcsc::Vector2D result_point = rcsc::inertia_n_step_point
                                        ( s.plan.getKickResultPos(),
                                          s.plan.getKickResultVel(),
                                          opp_team_access_step,
                                          param.ballDecay() );

    if ( result_point.x > param.theirTeamGoalLineX() )
    {
        double eval = result_point.x - std::fabs( result_point.y );

        agent->debugClient().addMessage( "ShootTest" );
        agent->debugClient().addLine( wm.ball().pos(), result_point );

        return std::make_pair( true, eval );
    }
    else
    {
        return std::make_pair( false, 0.0 );
    }
}
