// -*-c++-*-

/*!
  \file kick_table.cpp
  \brief kick planner and cache holder class Source File
*/

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public
 License as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version.

 This library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 Lesser General Public License for more details.

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "kick_simulator.h"

#include <rcsc/player/world_model.h>
#include <rcsc/geom/ray_2d.h>
#include <rcsc/geom/circle_2d.h>
#include <rcsc/geom/rect_2d.h>
#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/player_type.h>
#include <rcsc/game_time.h>
#include <rcsc/math_util.h>
#include <rcsc/soccer_math.h>
#include <rcsc/timer.h>

#include <algorithm>
#include <functional>

//#define DEBUG

namespace rcsc {

const double KickSimulator::NEAR_SIDE_RATE = 0.3;
const double KickSimulator::MID_RATE = 0.5;
const double KickSimulator::FAR_SIDE_RATE = 0.7;

struct TableCmp {
    bool operator()( const KickSimulator::Path & lhs,
                     const KickSimulator::Path & rhs )
      {
          if ( lhs.max_speed_ == rhs.max_speed_ )
          {
              return lhs.power_ < rhs.power_;
          }
          return lhs.max_speed_ > rhs.max_speed_;
      }
};

struct SequenceCmp {
    bool operator()( const KickSimulator::Sequence & lhs,
                     const KickSimulator::Sequence & rhs )
      {
          return lhs.score_ < rhs.score_;
      }
};

/*-------------------------------------------------------------------*/
/*!

 */
double
KickSimulator::calc_near_dist( const PlayerType & player_type )
{
    //       0.3 + 0.6*0.3 + 0.085 = 0.565
    // near: 0.3 + 0.7*0.3 + 0.085 = 0.595
    //       0.3 + 0.8*0.3 + 0.085 = 0.625
    //       0.3 + 0.9*0.3 + 0.085 = 0.655
    return bound( player_type.playerSize() + ServerParam::i().ballSize() + 0.1,
                  ( player_type.playerSize()
                    + ( player_type.kickableMargin() * NEAR_SIDE_RATE )
                    + ServerParam::i().ballSize() ),
                  player_type.kickableArea() - 0.2 );
}

/*-------------------------------------------------------------------*/
/*!

 */
double
KickSimulator::calc_mid_dist( const PlayerType & player_type )
{
    //      0.3 + 0.6*0.5 + 0.085 = 0.705
    // mid: 0.3 + 0.7*0.5 + 0.085 = 0.735
    //      0.3 + 0.8*0.5 + 0.085 = 0.765
    //      0.3 + 0.9*0.5 + 0.085 = 0.795
    return bound( player_type.playerSize() + ServerParam::i().ballSize() + 0.1,
                  ( player_type.playerSize()
                    + ( player_type.kickableMargin() * MID_RATE )
                    + ServerParam::i().ballSize() ),
                  player_type.kickableArea() - 0.2 );
}

/*-------------------------------------------------------------------*/
/*!

 */
double
KickSimulator::calc_far_dist( const PlayerType & player_type )
{
    //      0.3 + 0.6*0.75 + 0.085 = 0.88
    // far: 0.3 + 0.7*0.75 + 0.085 = 0.91
    //      0.3 + 0.8*0.75 + 0.085 = 0.94
    //      0.3 + 0.9*0.75 + 0.085 = 0.97
    return bound( player_type.playerSize() + ServerParam::i().ballSize() + 0.1,
                  ( player_type.playerSize()
                    + ( player_type.kickableMargin() * FAR_SIDE_RATE )
                    + ServerParam::i().ballSize() ),
                  player_type.kickableArea() - 0.2 );
}

/*-------------------------------------------------------------------*/
/*!

 */
Vector2D
KickSimulator::calc_max_velocity( const AngleDeg & target_angle,
                                  const double & krate,
                                  const Vector2D & ball_vel )
{
    const double ball_speed_max2 = std::pow( ServerParam::i().ballSpeedMax(), 2.0 );
    const double max_accel
        = std::min( ServerParam::i().maxPower() * krate,
                    ServerParam::i().ballAccelMax() );

    Ray2D desired_ray( Vector2D( 0.0, 0.0 ), target_angle );
    Circle2D next_reachable_circle( ball_vel, max_accel );

    Vector2D vel1, vel2;
    int num = next_reachable_circle.intersection( desired_ray, &vel1, &vel2 );

    if ( num == 0 )
    {
        return Vector2D( 0.0, 0.0 );
    }

    if ( num == 1 )
    {
        if ( vel1.r2() > ball_speed_max2 )
        {
            // next inertia ball point is within reachable circle.
            if ( next_reachable_circle.contains( Vector2D( 0.0, 0.0 ) ) )
            {
                // can adjust angle at least
                vel1.setLength( ServerParam::i().ballSpeedMax() );
            }
            else
            {
                // failed
                vel1.assign( 0.0, 0.0 );
            }
        }
        return vel1;
    }

    //
    // num == 2
    //   ball reachable circle does not contain the current ball pos.

    double length1 = vel1.r2();
    double length2 = vel2.r2();

    if ( length1 < length2 )
    {
        std::swap( vel1, vel2 );
        std::swap( length1, length2 );
    }

    if ( length1 > ball_speed_max2 )
    {
        if ( length2 > ball_speed_max2 )
        {
            // failed
            vel1.assign( 0.0, 0.0 );
        }
        else
        {
            vel1.setLength( ServerParam::i().ballSpeedMax() );
        }
    }

    return vel1;
}

/*-------------------------------------------------------------------*/
/*!

 */
int
KickSimulator::get_nearest_index( const Vector2D & current_self_pos,
                                  const AngleDeg & current_self_body,
                                  const PlayerType & self_type,
                                  const Vector2D & current_ball_pos )
{
    Vector2D rpos = current_ball_pos - current_self_pos;
    rpos.rotate( - current_self_body );
#if 0
    int min_index = -1;
    double min_dist2 = 100000.0;
    const std::vector< State >::const_iterator end = state_list.end();
    for ( std::vector< State >::const_iterator it = state_list.begin();
          it != end;
          ++it )
    {
        double d2 = it->pos_.dist2( rpos );
        if ( d2 < min_dist2 )
        {
            min_index = it->index_;
        }
    }
    return min_index;
#else
    const double near_dist = calc_near_dist( self_type );
    const double mid_dist = calc_mid_dist( self_type );
    const double far_dist = calc_far_dist( self_type );

    const double dist = rpos.r();
    const AngleDeg angle = rpos.th();

    const int dir_div = ( std::fabs( dist - near_dist ) < std::fabs( dist - mid_dist )
                          ? STATE_DIVS_NEAR
                          : std::fabs( dist - mid_dist ) < std::fabs( dist - far_dist )
                          ? STATE_DIVS_MID
                          : STATE_DIVS_FAR );

    int deg_idx = static_cast< int >( rint( dir_div * rint( angle.degree() + 180.0 ) / 360.0 ) );
    if ( deg_idx >= dir_div ) deg_idx = 0;

    return deg_idx;
#endif
}

/*-------------------------------------------------------------------*/
/*!

 */
KickSimulator &
KickSimulator::instance()
{
    static KickSimulator s_instance;
    return s_instance;
}

/*-------------------------------------------------------------------*/
/*!

 */
KickSimulator::KickSimulator()
{
    for ( int i = 0; i < MAX_DEPTH; ++ i )
    {
        M_state_cache[i].reserve( NUM_STATE );
    }

    createTables();
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::createTables()
{
    createStateList();

    MSecTimer timer;

    const double angle_step = 360.0 / DEST_DIR_DIVS;
    AngleDeg angle = -180.0;

    for ( int i = 0; i < DEST_DIR_DIVS; ++i, angle += angle_step )
    {
        createTable( angle, M_tables[i] );
    }

    dlog.addText( Logger::KICK,
                  __FILE__": createTables() elapsed %.3f [ms]",
                  timer.elapsedReal() );

#if 0
    const double kprate = ServerParam::i().kickPowerRate();
    for ( std::vector< State >::iterator s = M_state_list.begin();
          s != M_state_list.end();
          ++s )
    {
        std::cout << "  state "
                  << " index=" << s->index_
                  << " angle="  << s->pos_.th()
                  << " dist="  << s->pos_.r()
                  << " pos=" << s->pos_
                  << " kick_rate=" << s->kick_rate_
                  << " (" << s->kick_rate_ / kprate * 100.0 << "%)"
                  << std::endl;
    }

    for ( int i = 0; i < DEST_DIR_DIVS; ++i, angle += angle_step )
    {
        std::cout << "create table " << i << " : angle="  << angle << std::endl;
        for ( std::vector< Path >::iterator p = M_tables[i].begin();
              p != M_tables[i].end();
              ++p )
        {
            std::cout << "  table "
                      << " origin=" << p->origin_
                      << " dest=" << p->dest_
                      << " max_speed=" << p->max_speed_
                      << " power=" << p->power_
                      << std::endl;
        }
    }
#endif
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::createStateList()
{
    const ServerParam & param = ServerParam::i();

    const PlayerType player_type( param );

    const double near_dist = calc_near_dist( player_type );
    const double mid_dist = calc_mid_dist( player_type );
    const double far_dist = calc_far_dist( player_type );

    const double near_angle_step = 360.0 / STATE_DIVS_NEAR;
    const double mid_angle_step = 360.0 / STATE_DIVS_MID;
    const double far_angle_step = 360.0 / STATE_DIVS_FAR;

#if 0
    std::cout << "createStateList"
              << "\n  near_dist=" << near_dist
              << " mid_dist=" << mid_dist
              << " far_dist=" << far_dist
              << "\n  near_angle_step=" << near_angle_step
              << " mid_angle_step=" << mid_angle_step
              << " far_angle_step=" << far_angle_step
              << std::endl;
#endif
    int index = 0;
    M_state_list.clear();
    M_state_list.reserve( NUM_STATE );

    for ( int near = 0; near < STATE_DIVS_NEAR; ++near )
    {
        AngleDeg angle = -180.0 + ( near_angle_step * near );
        Vector2D pos = Vector2D::polar2vector( near_dist, angle );
        double krate = player_type.kickRate( param, near_dist, angle.degree() );
        M_state_list.push_back( State( index, pos, krate ) );
        ++index;
    }

    for ( int mid = 0; mid < STATE_DIVS_MID; ++mid )
    {
        AngleDeg angle = -180.0 + ( mid_angle_step * mid );
        Vector2D pos = Vector2D::polar2vector( mid_dist, angle );
        double krate = player_type.kickRate( param, mid_dist, angle.degree() );
        M_state_list.push_back( State( index, pos, krate ) );
        ++index;
    }

    for ( int far = 0; far < STATE_DIVS_FAR; ++far )
    {
        AngleDeg angle = -180.0 + ( far_angle_step * far );
        Vector2D pos = Vector2D::polar2vector( far_dist, angle );
        double krate = player_type.kickRate( param, far_dist, angle.degree() );
        M_state_list.push_back( State( index, pos, krate ) );
        ++index;
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::createTable( const AngleDeg & angle,
                            std::vector< Path > & table )
{
    const int max_combination = NUM_STATE * NUM_STATE;
    const int max_state = M_state_list.size();

    table.clear();
    table.reserve( max_combination );

    for ( int origin = 0; origin < max_state; ++origin )
    {
        for ( int dest = 0; dest < max_state; ++dest )
        {
            Vector2D vel = M_state_list[dest].pos_ - M_state_list[origin].pos_;
            Vector2D max_vel = calc_max_velocity( angle,
                                                  M_state_list[dest].kick_rate_,
                                                  vel );
            Vector2D accel = max_vel - vel;

            Path path( origin, dest );
            path.max_speed_ = max_vel.r();
            path.power_ = accel.r() / M_state_list[dest].kick_rate_;
            table.push_back( path );
        }
    }

    std::sort( table.begin(), table.end(), TableCmp() );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::updateState( const WorldModel & world )
{
    static GameTime s_update_time( -1, 0 );

    if ( s_update_time == world.time() )
    {
        return;
    }

    s_update_time = world.time();

    //
    // update current state
    //
    MSecTimer timer;
    createStateCache( world );
    dlog.addText( Logger::KICK,
                  __FILE__": updateState() elapsed %.3f [ms]",
                  timer.elapsedReal() );
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::createStateCache( const WorldModel & world )
{
#ifdef DEBUG
    dlog.addText( Logger::KICK,
                  __FILE__": createStateCache()" );
#endif

    const ServerParam & param = ServerParam::i();
    const PlayerType & self_type = world.self().playerType();
    const double near_dist = calc_near_dist( self_type );
    const double mid_dist = calc_mid_dist( self_type );
    const double far_dist = calc_far_dist( self_type );

    //
    // current state
    //
    {
        Vector2D rpos = world.ball().rpos();
        rpos.rotate( - world.self().body() );

        const double dist = rpos.r();
        const AngleDeg angle = rpos.th();

        const int dir_div = ( std::fabs( dist - near_dist ) < std::fabs( dist - far_dist )
                              ? STATE_DIVS_NEAR
                              : STATE_DIVS_FAR );

        M_current_state.index_ = static_cast< int >( rint( dir_div * rint( angle.degree() + 180.0 ) / 360.0 ) );
        if ( M_current_state.index_ >= dir_div ) M_current_state.index_ = 0;

        M_current_state.pos_ = world.ball().rpos();
        M_current_state.kick_rate_ = world.self().kickRate();
#ifdef DEBUG
        dlog.addText( Logger::KICK,
                      "__ current_state pos=(%.2f %.2f) kick_rate=%.3f",
                      world.ball().pos().x, world.ball().pos().y,
                      M_current_state.kick_rate_ );
#endif
        checkInterfereAt( world, 0, M_current_state );
    }

    //
    // create future state
    //

    Vector2D self_pos = world.self().pos();
    Vector2D self_vel = world.self().vel();

    for ( int i = 0; i < MAX_DEPTH; ++i )
    {
        M_state_cache[i].clear();

        self_pos += self_vel;
        self_vel *= self_type.playerDecay();

        int index = 0;
        for ( int near = 0; near < STATE_DIVS_NEAR; ++near )
        {
            Vector2D pos = M_state_list[index].pos_;
            double krate = self_type.kickRate( param, near_dist, pos.th().degree() );

            pos.rotate( world.self().body() );
            pos.setLength( near_dist );
            pos += self_pos;

            M_state_cache[i].push_back( State( index, pos, krate ) );
            checkInterfereAt( world, i + 1, M_state_cache[i].back() );
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "__ cache_near_%d index=%d pos=(%.2f %.2f) kick_rate=%f/%f",
                          i+1, index,
                          pos.x, pos.y,
                          krate, M_state_list[index].kick_rate_ );
#endif
            ++index;
        }

        for ( int mid = 0; mid < STATE_DIVS_MID; ++mid )
        {
            Vector2D pos = M_state_list[index].pos_;
            double krate = self_type.kickRate( param, mid_dist, pos.th().degree() );

            pos.rotate( world.self().body() );
            pos.setLength( mid_dist );
            pos += self_pos;

            M_state_cache[i].push_back( State( index, pos, krate ) );
            checkInterfereAt( world, i + 1, M_state_cache[i].back() );
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "__ cache_mid_%d index=%d pos=(%.2f %.2f) kick_rate=%f/%f",
                          i+1, index,
                          pos.x, pos.y,
                          krate,  M_state_list[index].kick_rate_ );
#endif
            ++index;
        }

        for ( int far = 0; far < STATE_DIVS_FAR; ++far )
        {
            Vector2D pos = M_state_list[index].pos_;
            double krate = self_type.kickRate( param, far_dist, pos.th().degree() );

            pos.rotate( world.self().body() );
            pos.setLength( far_dist );
            pos += self_pos;

            M_state_cache[i].push_back( State( index, pos, krate ) );
            checkInterfereAt( world, i + 1, M_state_cache[i].back() );
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "__ cache_far_%d index=%d pos=(%.2f %.2f) kick_rate=%f/%f",
                          i+1, index,
                          pos.x, pos.y,
                          krate, M_state_list[index].kick_rate_ );
#endif
            ++index;
        }

    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::checkCollisionAfterRelease( const WorldModel & world,
                                           const Vector2D & target_point,
                                           const double & first_speed )
{
#ifdef DEBUG
    dlog.addText( Logger::KICK,
                  __FILE__": checkCollisionAfterRelease()" );
#endif

    const PlayerType & self_type = world.self().playerType();

    const double collide_dist2 = std::pow( self_type.playerSize() + ServerParam::i().ballSize(), 2.0 );

    Vector2D self_pos = world.self().pos();
    Vector2D self_vel = world.self().vel();

    // check the release kick from current state

    self_pos += self_vel;
    self_vel *= self_type.playerDecay();

    {
        Vector2D release_pos = ( target_point - M_current_state.pos_ );
        release_pos.setLength( first_speed );

        if ( self_pos.dist2( release_pos ) < collide_dist2 )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "__ collision current_state self_pos=(%.2f %.2f) release_pos=(%.2f %.2f) dist=%.3f",
                          self_pos.x, self_pos.y,
                          release_pos.x, release_pos.y,
                          self_pos.dist( release_pos ) );
#endif
            M_current_state.flag_ |= SELF_COLLISION;
        }
        else
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "__ no collision with current_state" );
#endif
            M_current_state.flag_ &= ~SELF_COLLISION;
        }
    }

    // check the release kick from future state

    for ( int i = 0; i < MAX_DEPTH; ++ i )
    {
        self_pos += self_vel;
        self_vel *= self_type.playerDecay();

        const std::vector< State >::iterator end = M_state_cache[i].end();
        for ( std::vector< State >::iterator it = M_state_cache[i].begin();
              it != end;
              ++it )
        {
            Vector2D release_pos = ( target_point - it->pos_ );
            release_pos.setLength( first_speed );

            if ( self_pos.dist2( release_pos ) < collide_dist2 )
            {
#ifdef DEBUG
                dlog.addText( Logger::KICK,
                              "__ collision cached_state (%d) index=%d state_pos=(%.2f %.2f)"
                              " release_pos=(%.2f %.2f) dist=%.3f",
                              i + 1,
                              it->index_,
                              it->pos_.x, it->pos_.y,
                              release_pos.x, release_pos.y,
                              self_pos.dist( release_pos ) );
#endif
                it->flag_ |= SELF_COLLISION;
            }
            else
            {
#ifdef DEBUG
                dlog.addText( Logger::KICK,
                              "__ no collision cached_state (%d) index=%d (%.2f %.2f)",
                              i + 1,
                              it->index_,
                              it->pos_.x, it->pos_.y );
#endif
                it->flag_ &= ~SELF_COLLISION;
            }
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::checkInterfereAt( const WorldModel & world,
                                 const int cycle,
                                 State & state )
{
    static const Rect2D penalty_area( ServerParam::i().theirPenaltyAreaLineX(),
                                      ServerParam::i().penaltyAreaHalfWidth(),
                                      ServerParam::i().penaltyAreaLength(),
                                      ServerParam::i().penaltyAreaWidth() );

    int flag = 0x0000;

    const PlayerPtrCont::const_iterator end = world.opponentsFromBall().end();
    for ( PlayerPtrCont::const_iterator o = world.opponentsFromBall().begin();
          o != end;
          ++o )
    {
        if ( (*o)->isGhost() ) continue;
        if ( (*o)->posCount() >= 10 ) continue;
        if ( (*o)->distFromBall() > 10.0 ) continue;

        const double control_area = ( ( (*o)->goalie()
                                        && penalty_area.contains( (*o)->pos() )
                                        && penalty_area.contains( state.pos_ ) )
                                      ? ServerParam::i().catchableArea()
                                      : (*o)->playerTypePtr()->kickableArea() );
        const Vector2D opp_next = (*o)->pos() + (*o)->vel();
        const double opp_dist = opp_next.dist( state.pos_ );

        //
        // check kick possibility
        //

        if ( ! (*o)->isGhost()
             && (*o)->posCount() <= 2
             && opp_dist < control_area + 0.1 )
        {
            flag |= KICKABLE;
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "____ state %d (%.2f %.2f) kickable opp %d(%.1f %.1f)",
                          state.index_,
                          state.pos_.x, state.pos_.y,
                          (*o)->unum(),
                          (*o)->pos().x, (*o)->pos().y );
#endif
            continue;
        }

        //
        //
        //
        const AngleDeg opp_body =  ( (*o)->bodyCount() <= 1
                                     ? (*o)->body()
                                     : ( state.pos_ - opp_next ).th() );
        //
        // check tackle possibility
        //
        {
            Vector2D player_2_pos = state.pos_ - opp_next;
            player_2_pos.rotate( - opp_body );

            double tackle_dist = ( player_2_pos.x > 0.0
                                   ? ServerParam::i().tackleDist()
                                   : ServerParam::i().tackleBackDist() );
            if ( tackle_dist > 1.0e-5 )
            {
                double tackle_prob = ( std::pow( player_2_pos.absX() / tackle_dist,
                                                 ServerParam::i().tackleExponent() )
                                       + std::pow( player_2_pos.absY() / ServerParam::i().tackleWidth(),
                                                   ServerParam::i().tackleExponent() ) );
                if ( tackle_prob < 1.0
                     && 1.0 - tackle_prob > 0.7 ) // success probability
                {
                    flag |= TACKLABLE;
#ifdef DEBUG
                    dlog.addText( Logger::KICK,
                                  "____ state %d (%.2f %.2f) tackle opp %d(%.1f %.1f)",
                                  state.index_,
                                  state.pos_.x, state.pos_.y,
                                  (*o)->unum(),
                                  (*o)->pos().x, (*o)->pos().y );
#endif
                }
            }
        }

        // check kick or tackle possibility after dash

        const PlayerType * player_type = (*o)->playerTypePtr();
        const double max_accel
            = ServerParam::i().maxPower()
            * player_type->dashPowerRate()
            * player_type->effortMax();

        Vector2D player_next_2_pos = state.pos_ - opp_next;
        player_next_2_pos.rotate( - opp_body );

        if ( player_next_2_pos.absY() < control_area
             && ( player_next_2_pos.absX() < max_accel
                  || ( player_next_2_pos + Vector2D( max_accel, 0.0 ) ).r() < control_area + 0.1
                  || ( player_next_2_pos - Vector2D( max_accel, 0.0 ) ).r() < control_area + 0.1 )
             )
        {
            flag |= NEXT_KICKABLE;
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "____ state %d (%.2f %.2f) next kickable opp %d(%.1f %.1f)",
                          state.index_,
                          state.pos_.x, state.pos_.y,
                          (*o)->unum(),
                          (*o)->pos().x, (*o)->pos().y );
#endif
        }
        else if ( player_next_2_pos.absY() < ServerParam::i().tackleWidth() * 0.7
                  && player_next_2_pos.x - max_accel < ServerParam::i().tackleDist() - 0.1 )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "____ state %d (%.2f %.2f) next tackle opp %d(%.1f %.1f)",
                          state.index_,
                          state.pos_.x, state.pos_.y,
                          (*o)->unum(),
                          (*o)->pos().x, (*o)->pos().y );
#endif
            flag |= NEXT_TACKLABLE;
        }
    }

    state.flag_ = flag;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::checkInterfereAfterRelease( const WorldModel & world,
                                           const Vector2D & target_point,
                                           const double & first_speed )
{
    checkInterfereAfterRelease( world, target_point, first_speed, 0, M_current_state );

    for ( int i = 0; i < MAX_DEPTH; ++ i )
    {
        const std::vector< State >::iterator end = M_state_cache[i].end();
        for ( std::vector< State >::iterator state = M_state_cache[i].begin();
              state != end;
              ++state )
        {
            state->flag_ &= ~RELEASE_INTERFERE;

            checkInterfereAfterRelease( world, target_point, first_speed, i + 1, *state );
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::checkInterfereAfterRelease( const WorldModel & world,
                                           const Vector2D & target_point,
                                           const double & first_speed,
                                           const int cycle,
                                           State & state )
{
    static const Rect2D penalty_area( ServerParam::i().theirPenaltyAreaLineX(),
                                      - ServerParam::i().penaltyAreaHalfWidth(),
                                      ServerParam::i().penaltyAreaLength(),
                                      ServerParam::i().penaltyAreaWidth() );

    const PlayerPtrCont::const_iterator o_end = world.opponentsFromBall().end();
    for ( PlayerPtrCont::const_iterator o = world.opponentsFromBall().begin();
          o != o_end;
          ++o )
    {
        if ( (*o)->posCount() >= 10 ) continue;
        if ( (*o)->isGhost() ) continue;
        if ( (*o)->distFromBall() > 10.0 ) break;

        Vector2D opp_pos = (*o)->inertiaPoint( cycle );
        if ( ! opp_pos.valid() )
        {
            opp_pos = (*o)->pos() + (*o)->vel();
        }

        Vector2D ball_pos = state.pos_ - target_point;
        ball_pos.setLength( first_speed );
        ball_pos += state.pos_;

        double control_area2 = ( ( (*o)->goalie()
                                   && penalty_area.contains( opp_pos )
                                   && penalty_area.contains( ball_pos ) )
                                 ? ServerParam::i().catchableArea()
                                 : (*o)->playerTypePtr()->kickableArea() );
        control_area2 += 0.2;
        control_area2 *= control_area2;

        if ( ball_pos.dist2( opp_pos ) < control_area2 )
        {
            state.flag_ |= RELEASE_INTERFERE;
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "____ state %d-%d (%.2f %.2f) opp %d(%.1f %.1f) maybe interfere after release",
                          cycle,
                          state.index_,
                          state.pos_.x, state.pos_.y,
                          (*o)->unum(),
                          (*o)->pos().x, (*o)->pos().y );
#endif
            break;
        }
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
KickSimulator::simulateOneStep( const WorldModel & world,
                                const Vector2D & target_point,
                                const double & first_speed )
{
    if ( M_current_state.flag_ & SELF_COLLISION )
    {
        dlog.addText( Logger::KICK,
                      "xx__ 1 step: self collision" );
        return false;
    }

    if ( M_current_state.flag_ & RELEASE_INTERFERE )
    {
        dlog.addText( Logger::KICK,
                      "xx__ 1 step: opponent can interfere after release" );
        return false;
    }

    const double current_max_accel = std::min( M_current_state.kick_rate_ * ServerParam::i().maxPower(),
                                               ServerParam::i().ballAccelMax() );
    Vector2D target_vel = ( target_point - world.ball().pos() );
    target_vel.setLength( first_speed );

    Vector2D accel = target_vel - world.ball().vel();
    double accel_r = accel.r();
    if ( accel_r > current_max_accel )
    {
#ifdef DEBUG
        dlog.addText( Logger::KICK,
                      "xx__ 1 step: failed. max_vel=required_accel=%f > max_accel=%f",
                      accel_r, current_max_accel );
#endif
        Vector2D max_vel = calc_max_velocity( target_vel.th(),
                                              M_current_state.kick_rate_,
                                              world.ball().vel() );
        accel = max_vel - world.ball().vel();
        M_candidates.push_back( Sequence() );
        M_candidates.back().flag_ = M_current_state.flag_;
        M_candidates.back().pos_list_.push_back( world.ball().pos() + max_vel );
        M_candidates.back().speed_ = max_vel.r();
        M_candidates.back().power_ = accel.r() / M_current_state.kick_rate_;
        return false;
    }

    M_candidates.push_back( Sequence() );
    M_candidates.back().flag_ = M_current_state.flag_;
    M_candidates.back().pos_list_.push_back( world.ball().pos() + target_vel );
    M_candidates.back().speed_ = first_speed;
    M_candidates.back().power_ = accel_r / M_current_state.kick_rate_;
#ifdef DEBUG
    dlog.addText( Logger::KICK,
                  "ok__ 1 step: target_vel=(%.2f %.2f) required_accel=%.3f < max_accel=%.3f"
                  " kick_rate=%f power=%.1f",
                  target_vel.x, target_vel.y,
                  accel_r,
                  M_current_state.kick_rate_,
                  M_candidates.back().power_ );
#endif
    return true;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
KickSimulator::simulateTwoStep( const WorldModel & world,
                                const Vector2D & target_point,
                                const double & first_speed )
{
    static const double max_power = ServerParam::i().maxPower();
    static const double accel_max = ServerParam::i().ballAccelMax();
    static const double ball_decay = ServerParam::i().ballDecay();

    const PlayerType & self_type = world.self().playerType();

    const double current_max_accel = std::min( M_current_state.kick_rate_ * max_power, accel_max );

    const Vector2D my_final_pos
        = world.self().pos()
        + world.self().vel()
        + world.self().vel() * self_type.playerDecay();

    int success_count = 0;
    double max_speed2 = 0.0;

    for ( int i = 0; i < NUM_STATE; ++i )
    {
        const State & state = M_state_cache[0][i];

        if ( state.flag_ & KICKABLE )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 2 step: skip. exist kicable opp. state_pos=(%.2f %.2f)",
                          state.pos_.x, state.pos_.y );
#endif
            continue;
        }

        if ( state.flag_ & SELF_COLLISION )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 2 step: skip. self collision. state_pos=(%.2f %.2f)",
                          state.pos_.x, state.pos_.y );
#endif
            continue;
        }

        if ( state.flag_ & RELEASE_INTERFERE )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 2 step: interfere after release. state_pos=(%.2f %.2f)",
                          state.pos_.x, state.pos_.y );
#endif
            return false;
        }

        const Vector2D target_vel = ( target_point - state.pos_ ).setLengthVector( first_speed );

        Vector2D vel = state.pos_ - world.ball().pos();
        Vector2D accel = vel - world.ball().vel();
        double accel_r = accel.r();

        if ( accel_r > current_max_accel )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 2 step: failed(1) required_accel=%.3f > max_accel=%.3f",
                          accel_r, current_max_accel );
#endif
            continue;
        }
        vel *= ball_decay;

        accel = target_vel - vel;
        accel_r = accel.r();

        if ( accel_r > std::min( state.kick_rate_ * max_power, accel_max ) )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 2step: failed(2) required_accel=%.3f > max_accel=%.3f",
                          accel_r, std::min( state.kick_rate_ * max_power, accel_max ) );
#endif
            if ( success_count == 0 )
            {
                Vector2D max_vel = calc_max_velocity( target_vel.th(),
                                                      state.kick_rate_,
                                                      vel );
                double d2 = max_vel.r2();
                if ( max_speed2 < d2 )
                {
                    if ( max_speed2 == 0.0 )
                    {
                        M_candidates.push_back( Sequence() );
                    }
                    max_speed2 = d2;
                    accel = max_vel - vel;

                    M_candidates.back().flag_ = ( ( M_current_state.flag_ & ~RELEASE_INTERFERE )
                                                  | state.flag_ );
                    M_candidates.back().pos_list_.clear();
                    M_candidates.back().pos_list_.push_back( state.pos_ );
                    M_candidates.back().pos_list_.push_back( state.pos_ + max_vel );
                    M_candidates.back().speed_ = std::sqrt( max_speed2 );
                    M_candidates.back().power_ = accel.r() / state.kick_rate_;
#ifdef DEBUG
                    dlog.addText( Logger::KICK,
                                  "____ update max vel (%.2f %.2f) %.3f",
                                  max_vel.x, max_vel.y,
                                  M_candidates.back().speed_ );
#endif
                }
            }
            continue;
        }

        M_candidates.push_back( Sequence() );
        M_candidates.back().flag_ = ( ( M_current_state.flag_ & ~RELEASE_INTERFERE )
                                      | state.flag_ );
        M_candidates.back().pos_list_.push_back( state.pos_ );
        M_candidates.back().pos_list_.push_back( state.pos_ + target_vel );
        M_candidates.back().speed_ = first_speed;
        M_candidates.back().power_ = accel_r / state.kick_rate_;
#ifdef DEBUG
        dlog.addText( Logger::KICK,
                      "ok__ 2 step: last_power=%.2f subtarget=(%.2f %.2f)",
                      M_candidates.back().power_,
                      state.pos_.x, state.pos_.y );
#endif
    }

    return false;
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
KickSimulator::simulateThreeStep( const WorldModel & world,
                                  const Vector2D & target_point,
                                  const double & first_speed )
{
    const int max_loop = 256;

    static const double max_power = ServerParam::i().maxPower();
    static const double accel_max = ServerParam::i().ballAccelMax();
    static const double ball_decay = ServerParam::i().ballDecay();

    const double current_max_accel2 = square( std::min( M_current_state.kick_rate_ * max_power,
                                                        accel_max ) );

    AngleDeg target_rel_angle = ( target_point - world.self().pos() ).th() - world.self().body();
    double angle_deg = target_rel_angle.degree() + 180.0;
    int target_angle_index = static_cast< int >( rint( DEST_DIR_DIVS * ( angle_deg / 360.0 ) ) );
    if ( target_angle_index >= DEST_DIR_DIVS ) target_angle_index = 0;

    dlog.addText( Logger::KICK,
                  __FILE__": simulateThreeStep() target angle index = %d ",
                  target_angle_index );

    const std::vector< Path > & table = M_tables[target_angle_index];

    int success_count = 0;
    double max_speed2 = 0.0;

    int count = 0;
    const std::vector< Path >::const_iterator end = table.end();
    for ( std::vector< Path >::const_iterator it = table.begin();
          it != end && count < max_loop && success_count <= 10;
          ++it, ++count )
    {
        const State & state_1st = M_state_cache[0][it->origin_];
        const State & state_2nd = M_state_cache[1][it->dest_];

        if ( state_1st.flag_ & KICKABLE )
        {
#if 0
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: skip. exist kicable opp. state_1st pos=(%.2f %.2f)",
                          state_1st.pos_.x, state_1st.pos_.y );
#endif
            continue;
        }

        if ( state_2nd.flag_ & KICKABLE )
        {
#if 0
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: skip. exist kicable opp. state_2nd pos=(%.2f %.2f)",
                          state_2nd.pos_.x, state_2nd.pos_.y );
#endif
            continue;
        }

        if ( state_2nd.flag_ & SELF_COLLISION )
        {
#if 0
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: skip. self collision. state_2nd_pos=(%.2f %.2f)",
                          state_2nd.pos_.x, state_2nd.pos_.y );
#endif
            continue;
        }

        if ( state_2nd.flag_ & RELEASE_INTERFERE )
        {
#if 0
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: interfere after release. state_pos=(%.2f %.2f)",
                          state.pos_.x, state.pos_.y );
#endif
            return false;
        }


        const Vector2D target_vel = ( target_point - state_2nd.pos_ ).setLengthVector( first_speed );

        Vector2D vel1 = state_1st.pos_ - world.ball().pos();
        Vector2D accel = vel1 - world.ball().vel();
        double accel_r2 = accel.r2();

        if ( accel_r2 > current_max_accel2 )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: failed(1) required_accel=%.3f > max_accel=%.3f",
                          std::sqrt( accel_r2 ), std::sqrt( current_max_accel2 ) );
#endif
            continue;
        }
        vel1 *= ball_decay;

        Vector2D vel2 = state_2nd.pos_ - state_1st.pos_;
        accel = vel2 - vel1;
        accel_r2 = accel.r2();

        if ( accel_r2 > square( std::min( state_1st.kick_rate_ * max_power, accel_max ) ) )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: failed(2) required_accel=%.3f > max_accel=%.3f",
                          std::sqrt( accel_r2 ),
                          std::min( state_1st.kick_rate_ * max_power, accel_max ) );
#endif
            continue;
        }
        vel2 *= ball_decay;

        accel = target_vel - vel2;
        accel_r2 = accel.r2();
        if ( accel_r2 > square( std::min( state_2nd.kick_rate_ * max_power, accel_max ) ) )
        {
#ifdef DEBUG
            dlog.addText( Logger::KICK,
                          "xx__ 3 step: failed(3) required_accel=%.3f > max_accel=%.3f",
                          std::sqrt( accel_r2 ),
                          std::min( state_2nd.kick_rate_ * max_power, accel_max ) );
#endif
            if ( success_count == 0 )
            {
                Vector2D max_vel = calc_max_velocity( target_vel.th(),
                                                      state_2nd.kick_rate_,
                                                      vel2 );
                double d2 = max_vel.r2();
                if ( max_speed2 < d2 )
                {
                    if ( max_speed2 == 0.0 )
                    {
                        M_candidates.push_back( Sequence() );
                    }
                    max_speed2 = d2;
                    accel = max_vel - vel2;

                    M_candidates.back().flag_ = ( ( M_current_state.flag_ & ~RELEASE_INTERFERE )
                                                  | ( state_1st.flag_ & ~RELEASE_INTERFERE )
                                                  | state_2nd.flag_ );
                    M_candidates.back().pos_list_.clear();
                    M_candidates.back().pos_list_.push_back( state_1st.pos_ );
                    M_candidates.back().pos_list_.push_back( state_2nd.pos_ );
                    M_candidates.back().pos_list_.push_back( state_2nd.pos_ + max_vel );
                    M_candidates.back().speed_ = std::sqrt( max_speed2 );
                    M_candidates.back().power_ = accel.r() / state_2nd.kick_rate_;

#ifdef DEBUG
                    dlog.addText( Logger::KICK,
                                  "____ update max vel (%.2f %.2f) %.3f",
                                  max_vel.x, max_vel.y,
                                  M_candidates.back().speed_ );
#endif
                }
            }
            continue;
        }

        M_candidates.push_back( Sequence() );
        M_candidates.back().flag_ = ( ( M_current_state.flag_ & ~RELEASE_INTERFERE )
                                      | ( state_1st.flag_ & ~RELEASE_INTERFERE )
                                      | state_2nd.flag_ );
        M_candidates.back().pos_list_.push_back( state_1st.pos_ );
        M_candidates.back().pos_list_.push_back( state_2nd.pos_ );
        M_candidates.back().pos_list_.push_back( state_2nd.pos_ + target_vel );
        M_candidates.back().speed_ = first_speed;
        M_candidates.back().power_ = std::sqrt( accel_r2 ) / state_2nd.kick_rate_;

#ifdef DEBUG
        dlog.addText( Logger::KICK,
                      "ok__ 3 step: last_power=%.2f sub1=(%.2f %.2f) sub2(%.2f %.2f)",
                      M_candidates.back().power_,
                      state_1st.pos_.x, state_1st.pos_.y,
                      state_2nd.pos_.x, state_2nd.pos_.y );
#endif
        ++success_count;
    }

    dlog.addText( Logger::KICK,
                  "simulateThreeKick() solution_size=%d",
                  success_count );
    if ( success_count == 0 )
    {
        dlog.addText( Logger::KICK,
                      "simulateThreeKick() max_speed%.3f",
                      std::sqrt( max_speed2 ) );
    }

    return success_count > 0;
}

/*-------------------------------------------------------------------*/
/*!

 */
void
KickSimulator::evaluate( const double & first_speed )
{
    dlog.addText( Logger::KICK,
                  __FILE__": evaluate() candidate size=%d",
                  (int)M_candidates.size() );

    const double power_thr1 = ServerParam::i().maxPower() * 0.8;
    const double power_thr2 = ServerParam::i().maxPower() * 0.5;

    const std::vector< Sequence >::iterator end = M_candidates.end();
    for ( std::vector< Sequence >::iterator it = M_candidates.begin();
          it != end;
          ++it )
    {
        it->score_ = 1000.0;

        if ( it->speed_ < first_speed )
        {
            it->score_ = 0.0;
            it->score_ -= ( first_speed - it->speed_ ) * 100.0;
            continue;
        }

        if ( it->flag_ & TACKLABLE ) it->score_ -= 400.0;

        if ( it->flag_ & NEXT_TACKLABLE ) it->score_ -= 250.0;

        if ( it->flag_ & NEXT_KICKABLE ) it->score_ -= 600.0;

        const int n_kick = it->pos_list_.size();
        if ( n_kick == 3 )
        {
            it->score_ -= 100.0;
        }
        else if ( n_kick == 2 )
        {
            it->score_ -= 50.0;
        }

        if ( it->power_ > power_thr1 )
        {
            it->score_ -= 100.0;
        }
        else if ( it->power_ > power_thr2 )
        {
            it->score_ -= 50.0;
        }

        it->score_ -= it->power_;

        dlog.addText( Logger::KICK,
                      "__ sequence score %.2f flag=%x n_kick=%d last_kick_power=%f",
                      it->score_,
                      it->flag_,
                      n_kick,
                      it->power_ );
    }
}

/*-------------------------------------------------------------------*/
/*!

 */
bool
KickSimulator::simulate( const WorldModel & world,
                         const Vector2D & target_point,
                         const double & first_speed,
                         const int max_step,
                         Sequence & sequence )
{
    MSecTimer timer;

    double target_speed = bound( 0.0,
                                 first_speed,
                                 ServerParam::i().ballSpeedMax() );

    dlog.addText( Logger::KICK,
                  __FILE__": simulate() start. target=(%.2f %.2f) speed=%.2f",
                  target_point.x, target_point.y,
                  target_speed );

    M_candidates.clear();

    updateState( world );

    checkCollisionAfterRelease( world,
                                target_point,
                                target_speed );
    checkInterfereAfterRelease( world,
                                target_point,
                                target_speed );

    if ( max_step >= 1
         && simulateOneStep( world,
                             target_point,
                             target_speed ) )
    {
        dlog.addText( Logger::KICK,
                      __FILE__": simulate() found 1 step" );
    }

    if ( max_step >= 2
         && simulateTwoStep( world,
                             target_point,
                             target_speed ) )
    {
        dlog.addText( Logger::KICK,
                      __FILE__": simulate() found 2 step" );
    }

    if ( max_step >= 3
         && simulateThreeStep( world,
                               target_point,
                               target_speed ) )
    {
        dlog.addText( Logger::KICK,
                      __FILE__": simulate() found 3 step" );
    }

    // TODO:
    // 4 steps simulation

    // TODO:
    // dynamic evaluator
    evaluate( target_speed );

    if ( M_candidates.empty() )
    {
        return false;
    }

    sequence = *std::max_element( M_candidates.begin(),
                                  M_candidates.end(),
                                  SequenceCmp() );
    dlog.addText( Logger::KICK,
                  __FILE__": simulate() result next_pos=(%.1f %.1f)  flag=%x n_kick=%d speed=%.2f score=%.2f",
                  sequence.pos_list_.front().x,
                  sequence.pos_list_.front().y,
                  sequence.flag_,
                  (int)sequence.pos_list_.size(),
                  sequence.speed_,
                  sequence.score_ );

    dlog.addText( Logger::KICK,
                  __FILE__": simulate() total_elapsed=%.4f [ms].",
                  timer.elapsedReal() );
    return sequence.speed_ >= target_speed - rcsc::EPS;
}

}
